import unittest
import os
from typing import Dict, List

from bibtex_linter.parser import EntryType, BibTeXEntry, split_entries, parse_bibtex_file


class TestBibTeXEntry(unittest.TestCase):
    def test_parse_field_value(self) -> None:
        test_cases = [
            ("{John Doe}", "John Doe"),
            ("{{John Doe}}", "John Doe"),
            ('"John Doe"', "John Doe"),
            ("{{Industrial Digital Twin Association e. V.}}", "Industrial Digital Twin Association e. V."),
            ("{\\LaTeX}", "\\LaTeX"),
            ("{\\url{https://example.com}}", "\\url{https://example.com}"),
            ("   {Example Title}   ,", "Example Title"),
            ("{Multiple     Spaces}", "Multiple     Spaces"),
            ("{  Trimmed   }", "Trimmed"),
            ('"Surrounded by quotes" ,', "Surrounded by quotes"),
            ("   2020   ,", "2020"),
            ("{12345}", "12345"),
            ("{This is a {nested} value}", "This is a {nested} value"),
            ("{{This is a {{nested}} value}}", "This is a {{nested}} value"),
            ("Some plain text", "Some plain text"),
            ('"\\textit{Italicized}"', "\\textit{Italicized}"),
            ("{This value, has, commas}", "This value, has, commas"),
            ('"Quoted, with, commas"', "Quoted, with, commas"),
            ("{\\href{https://a.com}{A}}", "\\href{https://a.com}{A}"),
            ("{{\\href{https://a.com}{A}}}", "\\href{https://a.com}{A}"),
        ]

        for raw_value, expected in test_cases:
            with self.subTest(raw_value=raw_value):
                result = BibTeXEntry._parse_field_value(raw_value)
                self.assertEqual(expected, result)

    def test_split_fields_basic(self) -> None:
        entry = """@article{doe2020,
  author = {John Doe},
  title = {A Study},
  year = {2020}
}"""
        expected = [
            "  author = {John Doe}",
            "  title = {A Study}",
            "  year = {2020}"
        ]
        result = BibTeXEntry._split_fields(entry)
        self.assertEqual(expected, result)

    def test_split_fields_with_trailing_comma_and_newline(self) -> None:
        entry = """@book{smith2021,
  author = {Jane Smith},
  title = {The Book of Testing},
  year = {2021},
}"""
        expected = [
            "  author = {Jane Smith}",
            "  title = {The Book of Testing}",
            "  year = {2021}"
        ]
        result = BibTeXEntry._split_fields(entry)
        self.assertEqual(expected, result)

    def test_split_fields_multiline_values(self) -> None:
        entry = """@misc{nested2022,
  author = {{Industrial Digital Twin Association e. V.}},
  url = {https://example.com},
  note = {Line 1
          Line 2}
}"""
        expected = [
            "  author = {{Industrial Digital Twin Association e. V.}}",
            "  url = {https://example.com}",
            "  note = {Line 1\n          Line 2}"
        ]
        result = BibTeXEntry._split_fields(entry)
        self.assertEqual(expected, result)

    def test_split_fields_with_extra_whitespace(self) -> None:
        entry = ("@misc{id123,  \n    "
                 "author    =    {Someone}  , \n    "
                 "title=   {  Extra Spaces   }   , \n    "
                 "year= {2023}    }")
        expected = [
            "    author    =    {Someone}  ",
            "    title=   {  Extra Spaces   }   ",
            "    year= {2023}    "
        ]
        result = BibTeXEntry._split_fields(entry)
        self.assertEqual(expected, result)

    def test_split_fields_with_linebreak_after_entry_type(self) -> None:
        entry = """@misc
{
id456,
  author = {Spaced Out},
  title = {Linebreak test}
}"""
        expected = [
            "  author = {Spaced Out}",
            "  title = {Linebreak test}"
        ]
        result = BibTeXEntry._split_fields(entry)
        self.assertEqual(expected, result)

    def test_split_fields_missing_open_brace(self) -> None:
        entry = "article, author = {John Doe}, title = {Oops}"
        with self.assertRaises(KeyError):
            BibTeXEntry._split_fields(entry)


class TestSplitEntries(unittest.TestCase):
    def test_single_entry(self) -> None:
        raw = """@article{key1,
  author = {John Doe},
  title = {Example},
  year = {2020}
}"""
        entries = split_entries(raw)
        self.assertEqual(1, len(entries))
        self.assertIn("key1", entries[0])

    def test_multiple_entries(self) -> None:
        raw = """@article{key1,
  author = {John Doe},
  title = {Example 1},
  year = {2020}
}

@book{key2,
  author = {Jane Doe},
  title = {Example 2},
  year = {2021}
}"""
        entries = split_entries(raw)
        self.assertEqual(2, len(entries))
        self.assertIn("key1", entries[0])
        self.assertIn("key2", entries[1])

    def test_entry_with_nested_braces(self) -> None:
        raw = """@misc{key3,
  note = {Something with {nested} braces}
}"""
        entries = split_entries(raw)
        self.assertEqual(1, len(entries))
        self.assertIn("nested", entries[0])

    def test_entry_with_line_breaks(self) -> None:
        raw = """@online{key4,
  author = {Someone},
  title = {Line
Break},
  year = {2022}
}"""
        entries = split_entries(raw)
        self.assertEqual(1, len(entries))
        self.assertIn("Line\nBreak", entries[0])

    def test_incomplete_entry(self) -> None:
        raw = """@article{key5,
  title = {Missing closing brace}
"""
        entries = split_entries(raw)
        self.assertEqual(len(entries), 0)


class TestParseBibtexFile(unittest.TestCase):
    def test_parse_all_entries(self) -> None:
        bib_path = os.path.join(os.path.dirname(__file__), "test_refs.bib")
        entries = parse_bibtex_file(bib_path)
        self.assertEqual(17, len(entries))

        expected_types = {
            EntryType.ARTICLE: 1,
            EntryType.CONFERENCE: 1,
            EntryType.ONLINE: 1,
            EntryType.TECH_REPORT: 1,
            EntryType.BOOK: 1,
            EntryType.MISC: 9,
            EntryType.STANDARD: 1,
            EntryType.IN_BOOK: 1,
            EntryType.IN_COLLECTION: 1,
        }

        for entry_type, expected_count in expected_types.items():
            with self.subTest(entry_type=entry_type):
                actual_count = sum(1 for e in entries if e.entry_type == entry_type)
                self.assertEqual(expected_count, actual_count)

    def test_entry_fields_and_values(self) -> None:
        expected_entries: List[Dict[str, EntryType | Dict[str, str]]] = [
            {
                "type": EntryType.ARTICLE,
                "fields": {
                    "author": "Tests basic article",
                    "title": "Standard field format",
                    "year": "2020"
                }
            },
            {
                "type": EntryType.CONFERENCE,
                "fields": {
                    "author": "Should map to CONFERENCE",
                    "title": "Using alias INPROCEEDINGS",
                    "year": "2021"
                }
            },
            {
                "type": EntryType.ONLINE,
                "fields": {
                    "author": "Should map to ONLINE",
                    "url": "https://example.com"
                }
            },
            {
                "type": EntryType.TECH_REPORT,
                "fields": {
                    "author": "Should map to TECH_REPORT",
                    "title": "Tech report via alias"
                }
            },
            {
                "type": EntryType.BOOK,
                "fields": {
                    "author": "Extra spaces around field and value",
                    "title": "Trimmed Title",
                    "year": "2023"
                }
            },
            {
                "type": EntryType.MISC,
                "fields": {
                    "note": "This has nested braces inside",
                    "comment": "But only outermost braces should be stripped"
                }
            },
            {
                "type": EntryType.MISC,
                "fields": {
                    "author": "Double brace test",
                    "title": "Another level of nesting"
                }
            },
            {
                "type": EntryType.MISC,
                "fields": {
                    "author": "Quoted Author",
                    "title": "Simple quoted title"
                }
            },
            {
                "type": EntryType.MISC,
                "fields": {
                    "note": "This value\nspans multiple\nlines"
                }
            },
            {
                "type": EntryType.MISC,
                "fields": {
                    "howpublished": r"\url{https://wrapped-url.com}"
                }
            },
            {
                "type": EntryType.MISC,
                "fields": {
                    "title": r"\LaTeX command in value"
                }
            },
            {
                "type": EntryType.MISC,
                "fields": {
                    "author": "Trailing Comma",
                    "title": "Should be OK"
                }
            },
            {
                "type": EntryType.MISC,
                "fields": {
                    "author": "No Trailing Comma"
                }
            },
            {
                "type": EntryType.MISC,
                "fields": {
                    "author": "Newlines and spacing\neverywhere",
                    "title": "Still valid"
                }
            },
            {
                "type": EntryType.STANDARD,
                "fields": {
                    "author": "Tests EntryType.STANDARD",
                    "title": "Formal standard ref"
                }
            },
            {
                "type": EntryType.IN_BOOK,
                "fields": {
                    "author": "Part of a book",
                    "title": "Chapter Title",
                    "booktitle": "Whole Book Title"
                }
            },
            {
                "type": EntryType.IN_COLLECTION,
                "fields": {
                    "author": "Self-contained part of a collection",
                    "title": "In Collection Title",
                    "booktitle": "Collection Name"
                }
            },
        ]

        bib_path = os.path.join(os.path.dirname(__file__), "test_refs.bib")
        parsed_entries = parse_bibtex_file(bib_path)

        for expected in expected_entries:
            with self.subTest(expected=expected["fields"]):
                # (2025-04-24, s-heppner)
                # We can safely ignore the mypy warning here, since we wrote the `expected_entries` this way just above.
                expected_fields: Dict[str, str] = expected["fields"]  # type: ignore
                match = next(
                    (e for e in parsed_entries if all(e.fields.get(k) == v for k, v in expected_fields.items())),
                    None
                )
                all_field_sets = [e.fields for e in parsed_entries]
                self.assertIsNotNone(
                    match,
                    f"Missing or incorrect entry for:\nExpected Fields: {expected['fields']}\n"
                    f"Parsed Entries:\n{all_field_sets}"
                )
                # (2025-04-24, s-heppner)
                # We can safely ignore the mypy warning here, since we already asserted that `match` is not `None` in
                # the line above.
                self.assertEqual(expected["type"], match.entry_type)  # type: ignore


if __name__ == "__main__":
    unittest.main()



from typing import List, Dict
import dataclasses
import enum
import re


class EntryType(enum.Enum):
    """
    A collection of entry types that the LaTeX `IEEEtran` offers. Note that these only include what I
    need at the moment. The full list can be found at:
    https://ctan.net/macros/latex/contrib/IEEEtran/bibtex/IEEEtran_bst_HOWTO.pdf
    """
    ARTICLE = "ARTICLE"  # A typical journal article
    CONFERENCE = "CONFERENCE"  # A typical conference paper. Alias to: `IN_PROCEEDINGS`
    ONLINE = "ONLINE"  # A reference on the internet. Alias to: `ELECTRONIC`
    BOOK = "BOOK"  # Referencing a whole book
    IN_BOOK = "IN_BOOK"  # Referencing a part of a book (chapters or pages)
    IN_COLLECTION = "IN_COLLECTION"  # Referencing a part of a book that has its own name
    STANDARD = "STANDARD"  # Used for proposed or formally published standards
    TECH_REPORT = "TECH_REPORT"  # Used for technical reports, or reports about standards. Compare to `STANDARD`!
    MISC = "MISC"  # Anything else that does not fit the above

    @classmethod
    def from_string(cls, s: str) -> 'EntryType':
        """
        Get the `EntryType` from the string. Can deal with common aliases.

        :raises: KeyError, if the given string does not correspond to one of the entry types
        """
        s = s.upper()
        str_to_entry_type_map: Dict[str, "EntryType"] = {
            "ARTICLE": EntryType.ARTICLE,
            "CONFERENCE": EntryType.CONFERENCE,
            "INPROCEEDINGS": EntryType.CONFERENCE,

            "BOOK": EntryType.BOOK,
            "INBOOK": EntryType.IN_BOOK,
            "INCOLLECTION": EntryType.IN_COLLECTION,
            "STANDARD": EntryType.STANDARD,
            "TECHREPORT": EntryType.TECH_REPORT,

            "ONLINE": EntryType.ONLINE,
            "ELECTRONIC": EntryType.ONLINE,

            "MISC": EntryType.MISC,
        }
        return str_to_entry_type_map[s]


@dataclasses.dataclass
class BibTeXEntry:
    """
    An entry in a BibTeX file

    :ivar entry_type: Type of the entry (e.g. `@misc`). See `EntryType` for details
    :ivar name: Name or ID of the entry. So basically what is here: `@misc{Name_or_ID,`
    :ivar fields: Fields of the entry, as a Dict mapping the field key (e.g. `author`) to its cleaned up value.

    Note:
      The field's key is transformed via `.lower()`, so you can always expect non-capitalized characters.

    Note:
       When parsing multi-line field values, the additional white spaces are removed, but the new line characters are
       kept. For example, this:
       ```
       @misc{multiline_field,
         note = {This value
                 spans multiple
                 lines}
       }
       ```
       will be parsed to: `{"note": "This value\nspans multiple\nlines"}`. For the implementation details, check out
       the `BibTeXEntry._parse_field_value` static method.
    """
    entry_type: EntryType
    name: str
    fields: Dict[str, str]

    @classmethod
    def from_string(cls, entry_string: str) -> "BibTeXEntry":
        entry_type_string: str = entry_string.split("{")[0].lstrip("@")
        entry_type = EntryType.from_string(entry_type_string)

        name: str = entry_string.split("{")[1].split(",")[0]
        raw_fields = cls._split_fields(entry_string)
        fields: Dict[str, str] = {}
        for raw_field in raw_fields:
            raw_key, raw_value = raw_field.split("=")
            # Clean up key and value
            key = raw_key.strip(" ").lower()
            value = cls._parse_field_value(raw_value)
            fields[key] = value

        return BibTeXEntry(
            entry_type=entry_type,
            name=name,
            fields=fields,
        )

    @staticmethod
    def _split_fields(entry_string: str) -> List[str]:
        """
        Extract the list of fields from the entry_string (same input as into `from_string`)
        """
        entry_string = entry_string.strip()

        # Validate the entry starts with something like @type{ID,
        if not re.match(r"^@\w+\s*{", entry_string):
            raise KeyError(f"Invalid BibTeX entry format:\n\n{entry_string}\n\n")

        # Clean up trailing junk
        entry_string = entry_string.rstrip("}").rstrip(",\n")

        # Find the first `{` after entry type and extract what's inside
        start = entry_string.find('{')
        if start == -1:
            raise KeyError(f"Could not split the fields of the entry:\n\n{entry_string}\n\n")

        entry_string = entry_string[start + 1:]

        # Normalize comma-space-newline to comma-newline
        # It is possible to have trailing white spaces on the line `author = {{John Doe}}, \n`,
        # which we need to remove before we can split the entry into each field.
        entry_string = re.sub(r",\s*\n", ",\n", entry_string)

        raw_fields = entry_string.split(",\n")
        if raw_fields:
            raw_fields.pop(0)  # remove entry ID

        return raw_fields

    @staticmethod
    def _parse_field_value(raw_value: str) -> str:
        """
        Parse and normalize a field value.

        This will remove brackets `{}`, double brackets `{{}}`, quotation marks `"` and all additional stuff like
        commas `,` and white spaces from the field value, while leaving the value itself intact.

        Note:
          This does not change the `\\url{}` or `\\href{}{}` constructs, if they exist in the value.

        :param raw_value: The raw string of the value
        :return: The normalized value
        """
        # First we remove the trailing comma, and any spaces before and after it.
        raw_value = raw_value.strip().rstrip(',').strip()

        # If the value is in double brackets `{{`, we remove them
        if raw_value.startswith('{{') and raw_value.endswith('}}'):
            return raw_value[2:-2].strip()

        # If the value is in single brackets `{`, we remove them
        if (raw_value.startswith('{') and raw_value.endswith('}')) or \
                (raw_value.startswith('"') and raw_value.endswith('"')):
            return raw_value[1:-1].strip()

        return raw_value


def split_entries(raw_content: str) -> List[str]:
    """
    Split a file containing one or more entries into substrings containing each only one entry to prepare them for
    further parsing

    :param raw_content: Single string with one or more entries
    :return: List of substrings containing one entry each
    """
    entries = []
    brace_count = 0
    current_entry = []
    inside_entry = False

    for line in raw_content.splitlines():
        line = line.strip()
        if line.startswith('@'):
            inside_entry = True
            brace_count = 0
            current_entry = [line]
            brace_count += line.count('{') - line.count('}')
        elif inside_entry:
            current_entry.append(line)
            brace_count += line.count('{') - line.count('}')
            if brace_count == 0:
                entries.append('\n'.join(current_entry))
                inside_entry = False

    return entries


def parse_bibtex_file(filename: str) -> List[BibTeXEntry]:
    """
    Parse a BibTeX file and return the list of parsed `BibTeXEntry`s
    """
    with open(filename, "r") as file:
        raw_entries: List[str] = split_entries(file.read())
        entries: List[BibTeXEntry] = []
        for raw_entry_string in raw_entries:
            entries.append(BibTeXEntry.from_string(raw_entry_string))
    return entries



from typing import List, Set

from bibtex_linter.parser import BibTeXEntry, EntryType
from bibtex_linter.verification import (
    linter_rule,
    check_required_fields,
    check_omitted_fields,
)


@linter_rule(entry_type=EntryType.ARTICLE)
def check_article(entry: BibTeXEntry) -> List[str]:
    """
    Check that the required fields for `EntryType.ARTICLE` are there and that there are no fields present, that would
    be omitted in the final compiled LaTeX documents.

    :param entry: The BibTeXEntry
    :return: A list of string descriptions of rule violations for this entry.
    """
    invariant_violations: List[str] = []
    required_fields: Set[str] = {
        "author",
        "title",
        "journal",
        "year",
    }
    omitted_fields: Set[str] = {
        "language",
        "number",
        "url",
    }
    invariant_violations.extend(check_required_fields(entry, required_fields))
    invariant_violations.extend(check_omitted_fields(entry, omitted_fields))
    return invariant_violations


@linter_rule(entry_type=EntryType.CONFERENCE)
def check_conference(entry: BibTeXEntry) -> List[str]:
    """
    Check that the required fields for `EntryType.CONFERENCE` are there and that there are no fields present, that would
    be omitted in the final compiled LaTeX documents.

    :param entry: The BibTeXEntry
    :return: A list of string descriptions of rule violations for this entry.
    """
    invariant_violations: List[str] = []
    required_fields: Set[str] = {
        "author",
        "title",
        "booktitle",
        "publisher",
        "year",
    }
    omitted_fields: Set[str] = {
        "intype",
        "language",
        "number",
        "paper",
        "type",
        "url",
    }
    invariant_violations.extend(check_required_fields(entry, required_fields))
    invariant_violations.extend(check_omitted_fields(entry, omitted_fields))
    return invariant_violations


@linter_rule(entry_type=EntryType.ONLINE)
def check_online(entry: BibTeXEntry) -> List[str]:
    """
    Check that the required fields for `EntryType.ONLINE` are there and that there are no fields present, that would
    be omitted in the final compiled LaTeX documents.

    :param entry: The BibTeXEntry
    :return: A list of string descriptions of rule violations for this entry.
    """
    invariant_violations: List[str] = []
    required_fields: Set[str] = {
        "author",
        "title",
        "howpublished",
        "year",
        "note",
    }
    omitted_fields: Set[str] = {
        "language",
        "organization",
        "address",
        "url",
    }
    if "note" not in entry.fields.keys():
        invariant_violations.append(f"Entry '{entry.name}' is of type 'online' and needs a field 'note' with the URL.")
        # Todo: In the future, we could actually check that it contains an URL
    invariant_violations.extend(check_required_fields(entry, required_fields))
    invariant_violations.extend(check_omitted_fields(entry, omitted_fields))
    return invariant_violations


@linter_rule(entry_type=EntryType.BOOK)
def check_book(entry: BibTeXEntry) -> List[str]:
    """
    Check that the required fields for `EntryType.BOOK` are there and that there are no fields present, that would
    be omitted in the final compiled LaTeX documents.

    :param entry: The BibTeXEntry
    :return: A list of string descriptions of rule violations for this entry.
    """
    invariant_violations: List[str] = []
    required_fields: Set[str] = {
        "author",
        "title",
        "publisher",
        "year",
    }
    omitted_fields: Set[str] = {
        "editor",
        "language",
        "volume",
        "number",
        "url",
    }
    invariant_violations.extend(check_required_fields(entry, required_fields))
    invariant_violations.extend(check_omitted_fields(entry, omitted_fields))
    return invariant_violations


@linter_rule(entry_type=EntryType.IN_BOOK)
def check_in_book(entry: BibTeXEntry) -> List[str]:
    """
    Check that the required fields for `EntryType.IN_BOOK` are there and that there are no fields present, that would
    be omitted in the final compiled LaTeX documents.

    :param entry: The BibTeXEntry
    :return: A list of string descriptions of rule violations for this entry.
    """
    invariant_violations: List[str] = []
    required_fields: Set[str] = {
        "author",
        "title",
        "publisher",
        "year",
    }
    omitted_fields: Set[str] = {
        "editor",
        "language",
        "number",
        "url",
    }
    if "chapter" not in entry.fields.keys() or "pages" not in entry.fields.keys():
        invariant_violations.append(f"Entry {entry.name} needs to contain one of the "
                                    f"following fields: [chapter, pages].")
    invariant_violations.extend(check_required_fields(entry, required_fields))
    invariant_violations.extend(check_omitted_fields(entry, omitted_fields))
    return invariant_violations


@linter_rule(entry_type=EntryType.IN_COLLECTION)
def check_in_collection(entry: BibTeXEntry) -> List[str]:
    """
    Check that the required fields for `EntryType.IN_COLLECTION` are there and that there are no fields present, that
    would be omitted in the final compiled LaTeX documents.

    :param entry: The BibTeXEntry
    :return: A list of string descriptions of rule violations for this entry.
    """
    invariant_violations: List[str] = []
    required_fields: Set[str] = {
        "author",
        "title",
        "booktitle",
        "publisher",
        "year",
    }
    omitted_fields: Set[str] = {
        "language",
        "number",
        "url",
    }
    invariant_violations.extend(check_required_fields(entry, required_fields))
    invariant_violations.extend(check_omitted_fields(entry, omitted_fields))
    return invariant_violations


@linter_rule(entry_type=EntryType.STANDARD)
def check_standard(entry: BibTeXEntry) -> List[str]:
    """
    Check that the required fields for `EntryType.STANDARD` are there and that there are no fields present, that would
    be omitted in the final compiled LaTeX documents.

    :param entry: The BibTeXEntry
    :return: A list of string descriptions of rule violations for this entry.
    """
    invariant_violations: List[str] = []
    required_fields: Set[str] = {
        "author",
        "title",
        "howpublished",
        "year",
        "note",
    }
    omitted_fields: Set[str] = {
        "language",
        "organization",
        "institution",
        "type",
        "number",
        "revision",
        "address",
        "url",
    }
    invariant_violations.extend(check_required_fields(entry, required_fields))
    invariant_violations.extend(check_omitted_fields(entry, omitted_fields))
    return invariant_violations


@linter_rule(entry_type=EntryType.TECH_REPORT)
def check_tech_report(entry: BibTeXEntry) -> List[str]:
    """
    Check that the required fields for `EntryType.TECH_REPORT` are there and that there are no fields present, that
    would be omitted in the final compiled LaTeX documents.

    :param entry: The BibTeXEntry
    :return: A list of string descriptions of rule violations for this entry.
    """
    invariant_violations: List[str] = []
    required_fields: Set[str] = {
        "author",
        "title",
        "institution",
        "year",
    }
    omitted_fields: Set[str] = {
        "language",
        "howpublished",
        "url",
    }
    if "howpublished" in entry.fields.keys():
        invariant_violations.append(f"Entry {entry.name} is of type 'techreport', which does not render field "
                                    f"'howpublished'. Either use field 'institution' instead, or switch to a different "
                                    f"entry type completely.")
    invariant_violations.extend(check_required_fields(entry, required_fields))
    invariant_violations.extend(check_omitted_fields(entry, omitted_fields))
    return invariant_violations


@linter_rule(entry_type=EntryType.MISC)
def check_misc(entry: BibTeXEntry) -> List[str]:
    """
    Check that the required fields for `EntryType.MISC` are there and that there are no fields present, that would
    be omitted in the final compiled LaTeX documents.

    :param entry: The BibTeXEntry
    :return: A list of string descriptions of rule violations for this entry.
    """
    invariant_violations: List[str] = []
    required_fields: Set[str] = {
        "author",
        "title",
        "howpublished",
        "year",
    }
    omitted_fields: Set[str] = {
        "language",
        "organization",
        "address",
        "pages",
        "url",
    }
    invariant_violations.extend(check_required_fields(entry, required_fields))
    invariant_violations.extend(check_omitted_fields(entry, omitted_fields))
    return invariant_violations



from typing import List
import argparse
import importlib.util
import sys

from bibtex_linter.verification import verify
from bibtex_linter.parser import BibTeXEntry, parse_bibtex_file


def import_from_path(file_path: str) -> None:
    """
    Import a given module using its path.
    """
    # (2025-04-24, s-heppner)
    # This is taken directly from the importlib documentation:
    # https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
    # It seems a bit cursed, but I guess as long as it works and really only used on known and safe `rules.py`...
    module_name: str = "ruleset"
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    if not spec or not spec.loader:
        raise ImportError(f"Could not import ruleset from '{file_path}'.")
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)


def main() -> None:
    parser = argparse.ArgumentParser(description="Verify a .bib file using a set of defined rules.")
    parser.add_argument("filepath", type=str, help="Path to the .bib file to verify")
    parser.add_argument("ruleset",
                        type=str,
                        nargs="?",
                        default=None,
                        help="Path to the rules.py that define the rules. If left empty, the default ruleset is used. "
                             "WARNING: Executes the Python code inside rules.py, so be sure that it's safe!")

    args = parser.parse_args()

    # Try to import the ruleset
    if args.ruleset is None:
        import bibtex_linter.default_rules
        print("Using the default ruleset.")
    else:
        print(f"Importing rules from {args.ruleset}.")
        import_from_path(args.ruleset)

    entries: List[BibTeXEntry] = parse_bibtex_file(args.filepath)
    had_violations = False
    total_number_of_violations: int = 0

    for entry in entries:
        violations: List[str] = verify(entry)
        total_number_of_violations += len(violations)
        if violations:
            had_violations = True
            print(f"\nEntry '{entry.name}' of type '{entry.entry_type.name}' failed verification:")
            print("  ❌ Invariant Violations:")
            for issue in violations:
                print(f"    - {issue}")

    print(f"\n\nFound {total_number_of_violations} invariant violation(s) in {len(entries)} entries.")

    if not had_violations:
        print("All entries passed verification.")
        sys.exit(0)  # Exit as success

    sys.exit(1)  # Exit as failure


if __name__ == "__main__":
    main()



"""
This module implements verification of constraints or invariants.

When using the decorators, they automatically load the method below them into the `_rules` list at time
of import.
"""
from typing import Callable, TypeVar, List, Optional, Set

from bibtex_linter.parser import BibTeXEntry, EntryType

# The dynamic list of known rules.
# This list gets updated when a method with the `@linter_rule` decorator gets imported.
_rules: List[Callable[[BibTeXEntry], List[str]]] = []

# For the type annotations, we define a `LINTER_RULE_TYPE` variable, which describes the type of the methods that
# define the linter rules.
LINTER_RULE_TYPE = TypeVar("LINTER_RULE_TYPE", bound=Callable[[BibTeXEntry], List[str]])


def linter_rule(entry_type: Optional[EntryType] = None) -> Callable[[LINTER_RULE_TYPE], LINTER_RULE_TYPE]:
    """
    Decorator to mark a method defines rules to be checked by the linter for a specific entry type.

    If `entry_type` is `None`, we assume it is valid for all types.
    """
    def wrapper(func: LINTER_RULE_TYPE) -> LINTER_RULE_TYPE:
        setattr(func, "_is_invariant", True)
        setattr(func, "_entry_type", entry_type)
        _rules.append(func)
        return func
    return wrapper


def check_required_fields(entry: BibTeXEntry, fields: Set[str]) -> List[str]:
    """
    Helper function to check the existence of a set of required fields for the given entry.
    """
    existing_fields: Set[str] = set(entry.fields.keys())
    if not fields.issubset(entry.fields):
        missing = fields - existing_fields
        return [f"Entry '{entry.name}' misses the following required fields: [{', '.join(sorted(missing))}]"]
    return []


def check_omitted_fields(entry: BibTeXEntry, fields: Set[str]) -> List[str]:
    """
    Helper function to check the existence of a set of omitted fields for the given entry.
    """
    existing_fields: Set[str] = set(entry.fields.keys())
    omitted_fields_present = fields & existing_fields

    if omitted_fields_present:
        return [f"Entry '{entry.name}' has fields present that would be omitted in the compiled document: "
                f"[{', '.join(sorted(omitted_fields_present))}]. This could lead to a loss of information."]
    return []


def verify(entry: BibTeXEntry) -> List[str]:
    """
    Call this function to execute all imported methods that have the `invariant` decorator.

    Warning: This is basically remote code execution, so be sure to know what methods are imported!
    """
    errors = []

    for check in _rules:
        entry_type = getattr(check, "_entry_type", None)
        if entry_type is None or entry_type == entry.entry_type:
            errors.extend(check(entry))

    return errors
